/*
 * Copyright (c) BindView Development Corporation, 2001
 * See LICENSE file.
 * Author: Todd Sabin <tsabin@razor.bindview.com>
 */


#define UNICODE
#define _UNICODE
#include <tchar.h>
#include <windows.h>
#include <aclui.h>
#include <aclapi.h>
#include <ntsecapi.h>

/*
 * These are not prototyped in normal headers, but they are exported
 * from advapi32.lib
 */
extern "C" {
DWORD WINAPI
LsaQuerySecurityObject (LSA_HANDLE h, SECURITY_INFORMATION si,
                        PSECURITY_DESCRIPTOR *ppSD);
DWORD WINAPI
LsaSetSecurityObject (LSA_HANDLE h, SECURITY_INFORMATION si,
                      PSECURITY_DESCRIPTOR pSD);
}

static SI_ACCESS g_lsaAccess[] = {
    { &GUID_NULL, POLICY_VIEW_LOCAL_INFORMATION, L"View Local Info", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_VIEW_AUDIT_INFORMATION, L"View Audit Info", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_GET_PRIVATE_INFORMATION, L"Get Private Info", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_TRUST_ADMIN, L"Administer Trusts", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_CREATE_ACCOUNT, L"Create Account", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_CREATE_SECRET, L"Create Secret", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_CREATE_PRIVILEGE, L"Create Privilege", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_SET_DEFAULT_QUOTA_LIMITS, L"Set Default Quota Limits", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_SET_AUDIT_REQUIREMENTS, L"Set Audit Requirements", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_AUDIT_LOG_ADMIN, L"Administer Logs", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_SERVER_ADMIN, L"Administer Server", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, POLICY_LOOKUP_NAMES, L"Lookup Names", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, DELETE, L"Delete", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, READ_CONTROL, L"Read Control", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, WRITE_DAC, L"Write DAC", SI_ACCESS_SPECIFIC },
    { &GUID_NULL, WRITE_OWNER, L"Write Owner", SI_ACCESS_SPECIFIC },

    /* Full Control is different between NT4 and W2K.  I'm putting both
       of these here for (hopeful) clarity */
    { &GUID_NULL, 0x000f0fff, L"Full Control (nt4)", SI_ACCESS_GENERAL },
    { &GUID_NULL, 0x000f1fff, L"Full Control (w2k)", SI_ACCESS_GENERAL },
    { &GUID_NULL, 0x00020801, L"Execute", SI_ACCESS_GENERAL },
};

static GENERIC_MAPPING g_lsaGenericMapping = {
    POLICY_READ,
    POLICY_WRITE,
    POLICY_EXECUTE,
    POLICY_ALL_ACCESS
};

static SI_INHERIT_TYPE g_lsaInheritTypes[] = {
    { &GUID_NULL,     0, L"This object only" },
};

static ULONG g_extra_flags;

struct CLsaObjectSecurityInfoBase : ISecurityInformation
{
    long m_cRefs;
    LSA_HANDLE m_handle;
    const wchar_t* m_pszObjectName;
    const wchar_t* m_pszServerName;
    const wchar_t* m_pszPageTitle;

    CLsaObjectSecurityInfoBase (LSA_HANDLE h,
                                const wchar_t* pszObjectName,
                                const wchar_t* pszServerName,
                                const wchar_t* pszPageTitle = 0)
        : m_cRefs (0),
          m_handle (h),
          m_pszObjectName (pszObjectName),
          m_pszServerName (pszServerName),
          m_pszPageTitle (pszPageTitle)  {}
    virtual ~CLsaObjectSecurityInfoBase () {}

    STDMETHODIMP QueryInterface (REFIID iid, void** ppv)
    {
        if ((IID_IUnknown == iid) || (IID_ISecurityInformation == iid)) {
            *ppv = static_cast<ISecurityInformation*>(this);
        } else {
            *ppv = 0;
            return E_NOINTERFACE;
        }
        reinterpret_cast<IUnknown*>(*ppv)->AddRef ();
        return S_OK;
    }
    STDMETHODIMP_ (ULONG) AddRef ()
    {
        return ++m_cRefs;
    }
    STDMETHODIMP_ (ULONG) Release ()
    {
        long n = --m_cRefs;
        if (n==0) {
            NTSTATUS rc = LsaClose (m_handle);
            delete this;
        }
        return n;
    }

    STDMETHODIMP GetObjectInformation (SI_OBJECT_INFO* poi)
    {
        poi->dwFlags = (SI_EDIT_OWNER | SI_EDIT_PERMS | SI_ADVANCED | SI_NO_ACL_PROTECT
                        | g_extra_flags);

        poi->hInstance      = GetModuleHandle (NULL);
        poi->pszServerName  = const_cast<wchar_t*> (m_pszServerName);
        poi->pszObjectName  = const_cast<wchar_t*> (m_pszObjectName);
        poi->pszPageTitle   = const_cast<wchar_t*> (m_pszPageTitle);

        if (m_pszPageTitle)
            poi->dwFlags |= SI_PAGE_TITLE;

        return S_OK;
    }

    STDMETHODIMP GetSecurity (SECURITY_INFORMATION ri, void** ppsd, BOOL bDefault)
    {
        DWORD rc = LsaQuerySecurityObject (m_handle, ri, ppsd);
        return rc ? HRESULT_FROM_WIN32 (rc) : S_OK;
    }

    STDMETHODIMP SetSecurity (SECURITY_INFORMATION ri, void* psd)
    {
        DWORD rc = LsaSetSecurityObject (m_handle, ri, psd);
        return rc ? HRESULT_FROM_WIN32 (rc) : S_OK;
    }

    STDMETHODIMP PropertySheetPageCallback (HWND hwnd, UINT msg, SI_PAGE_TYPE pt)
    {
        return S_OK;
    }
};

struct CLsaPolicySecurityInfo : CLsaObjectSecurityInfoBase
{
    CLsaPolicySecurityInfo (LSA_HANDLE hPol, const wchar_t* pszObjectTitle,
                            const wchar_t* pszServerName = 0,
                            const wchar_t* pszPageTitle = 0)
        : CLsaObjectSecurityInfoBase (hPol,
                                      pszObjectTitle,
                                      pszServerName,
                                      pszPageTitle)
    {}
    
    ~CLsaPolicySecurityInfo ()
    {}

    STDMETHODIMP GetAccessRights (const GUID*,
                                  DWORD dwFlags,
                                  SI_ACCESS** ppAccess,
                                  ULONG* pcAccesses,
                                  ULONG* piDefaultAccess)
    {
        *ppAccess = const_cast<SI_ACCESS*> (g_lsaAccess);
        *pcAccesses = sizeof (g_lsaAccess) / sizeof (g_lsaAccess[0]);
        *piDefaultAccess = 0;
        return S_OK;
    }

    STDMETHODIMP MapGeneric (const GUID*, UCHAR* pAceFlags, ACCESS_MASK* pMask)
    {
        MapGenericMask (pMask, const_cast<GENERIC_MAPPING*> (&g_lsaGenericMapping));
        return S_OK;
    }

    STDMETHODIMP GetInheritTypes (SI_INHERIT_TYPE** ppInheritTypes, ULONG* pcInheritTypes)
    {
        *ppInheritTypes = g_lsaInheritTypes;
        *pcInheritTypes = sizeof (g_lsaInheritTypes) / sizeof (g_lsaInheritTypes[0]);
        return S_OK;
    }

};

int
_tmain (int argc, TCHAR *argv[])
{
    LSA_HANDLE hLsa;
    NTSTATUS rc;
    LSA_OBJECT_ATTRIBUTES oa = { sizeof (oa) };
    LSA_UNICODE_STRING uni_target;
    LSA_UNICODE_STRING *target = NULL;
    wchar_t machine[100] = L"127.0.0.1";
    ULONG desired_access = MAXIMUM_ALLOWED;

    for (int argnum=1; argnum < argc; argnum++) {
        if (argv[argnum][0] == '-') {
            switch (argv[argnum][1]) {
            case 's':
                desired_access |= ACCESS_SYSTEM_SECURITY;
                g_extra_flags |= SI_EDIT_AUDITS;
                break;

            case 'p':
                g_extra_flags |= SI_SERVER_IS_DC;
                break;

            default:
                _tprintf (_T ("Unknown option: %s\n"), argv[argnum]);
                return 0;
                break;
            }
        } else {
            wcsncpy (machine, argv[argnum], sizeof (machine) / sizeof (machine[0]));
            machine[sizeof (machine) / sizeof (machine[0])-1] = L'\0';
            uni_target.Length = wcslen (machine) * 2;
            uni_target.MaximumLength = uni_target.Length + 2;
            uni_target.Buffer = machine;
            target = &uni_target;
        }
    }

    rc = LsaOpenPolicy (target, &oa, desired_access, &hLsa);
    if (rc < 0) {
        MessageBox (HWND_DESKTOP, L"Failed to open LSA Policy", L"Error", MB_OK);
        return 0;        
    }

    CLsaPolicySecurityInfo* psi =
        new CLsaPolicySecurityInfo (hLsa, L"Lsa Policy",
                                    target ? machine : NULL,
                                    L"Lsa Policy Object");
    psi->AddRef ();
    HPROPSHEETPAGE ps = CreateSecurityPage (psi);
    psi->Release ();
    psi = NULL;

    PROPSHEETHEADER psh = { sizeof (psh) };
    psh.hInstance = GetModuleHandle (NULL);
    psh.pszCaption = L"Lsa Policy Object Security Editor";
    psh.nPages = 1;
    psh.phpage = &ps;

    PropertySheet (&psh);

    return 0;
}
